from .tracker import Tracker


class NullTracker(Tracker):
    def __init__(self):
        pass

    def update(self, readings):
        """
        Does nothing.
        """
        return

    @property
    def values(self):
        """
        Returns {}.
        """
        return {}



"""
Creates a thread for each :mod:`Output <snsary.outputs.output>` that subscribes to. Any :mod:`Readings <snsary.models.reading>` are published to a separate queue for each :mod:`Output <snsary.outputs.output>`, with the thread logging any errors from relaying to it. Using AsyncStream helps avoid issues due to flakey or time-consuming :mod:`Outputs <snsary.outputs.output>`.
"""

from queue import Queue
from threading import Thread

from .stream import Stream


class AsyncStream(Stream):
    def __init__(self):
        self.__relays = {}

    def publish(self, reading):
        for queue in self.__relays.values():
            queue.put(reading)

    def subscribe(self, output):
        queue = Queue()
        self.__relays[output] = queue

        def _relay():
            while True:
                try:
                    output.publish(queue.get())
                except Exception as e:
                    output.logger.exception(e)

        Thread(target=_relay, daemon=True).start()



"""
Manages the :mod:`Monitors <snsary.utils.tracing.monitors>` associated with an instrumented block of code. Calling ``get`` will return a list of instances of any :mod:`Monitors <snsary.utils.tracing.monitors>` set in :mod:`Config <snsary.utils.tracing.config>` for the specified ``trace_id``: ::

    config = Config()
    config.set("some.id.monitors", [Monitor.factory()])

    registry = Registry(config)
    registry.get("some.id")  # => [<Monitor>]

The same instances are returned across calls to ``get`` for the same ``trace_id``. Setting ``thread_aware`` in :mod:`Config <snsary.utils.tracing.config>` means the returned instances will also be unique across threads.
"""

from threading import current_thread


class Registry:
    def __init__(self, config):
        self.__backend = {}
        self.__config = config

    def reset(self):
        self.__backend = {}

    def get(self, trace_id):
        registry_id = self.__registry_id(trace_id)

        if registry_id not in self.__backend:
            monitors = self.__monitors(trace_id)
            self.__backend[registry_id] = monitors

        return self.__backend[registry_id]

    def __monitors(self, trace_id):
        return [
            monitor_class()
            for monitor_class in self.__config.get(
                f"{trace_id}.monitors",
                default=[],
            )
        ]

    def __registry_id(self, trace_id):
        registry_id = f"[trace_id={trace_id}]"

        thread_aware = self.__config.get(
            f"{trace_id}.thread_aware",
            default=False,
        )

        if thread_aware:
            thread_ident = current_thread().ident
            registry_id += f"[thread_ident={thread_ident}]"

        return registry_id



"""
Publishes zero or more :mod:`Readings <snsary.models.Reading>` depending on what the specified :mod:`function <snsary.functions>` returns for a given :mod:`Reading <snsary.models.reading>`. This could be nothing, the same reading, or multiple, different readings.
"""
from wrapt import synchronized

from .simple_stream import SimpleStream


class FuncStream(SimpleStream):
    def __init__(
        self,
        stream,
        function=lambda reading: [reading],
    ):
        SimpleStream.__init__(self)
        stream.subscribe(self)
        self.__function = function

    @synchronized
    def publish(self, reading):
        output_readings = self.__function(reading)

        for output_reading in output_readings:
            SimpleStream.publish(self, output_reading)



"""
Depending on the output, it may be more efficient to dispatch multiple :mod:`Readings<snsary.models.reading>` together. This can be done by inheriting from BatchOutput, which requires a ``publish_batch`` method.

BatchOutput protects against concurrent execution of its ``publish`` method. BatchOutput is also a :mod:`Service <snsary.system.service>` and will try to publish any remaining Readings when told to ``stop()``.
"""

from datetime import datetime

from wrapt import synchronized

from snsary import system
from snsary.utils import tracing

from .output import Output


class BatchOutput(Output, system.Service):
    def __init__(self, *, max_size=100, max_wait_seconds=10):
        Output.__init__(self)
        self.__max_size = max_size
        self.__readings = []
        self.__max_wait_seconds = max_wait_seconds
        self.__last_publish = datetime.utcnow().timestamp()

    @tracing.capture_exceptions("snsary.outputs.batch_output")
    def flush(self):
        # in case of multiple consecutive batch/age flushes
        if not self.__readings:
            return

        self.logger.info(f"Sending {len(self.__readings)} readings.")
        self.publish_batch(self.__readings)
        self.__readings = []

    def stop(self):
        if self.__readings:
            self.flush()

    @synchronized
    def publish(self, reading):
        self.__readings += [reading]
        self.__try_publish_large_batch()
        self.__try_publish_old_batch()

    def publish_batch(self, readings):
        raise NotImplementedError()

    def __try_publish_large_batch(self):
        if len(self.__readings) < self.__max_size:
            return

        self.flush()

    def __try_publish_old_batch(self):
        now = datetime.utcnow().timestamp()
        delay = now - self.__last_publish

        if delay > self.__max_wait_seconds:
            self.flush()
            self.__last_publish = now



"""
Wrapper for the Adafruit SGP30 sensor, inheriting from :mod:`GenericSensor <snsary.contrib.adafruit.generic>`:

    - Waits 15 seconds for the device to warm up (as per the `datasheet <https://www.sensirion.com/fileadmin/user_upload/customers/sensirion/Dokumente/9_Gas_Sensors/Datasheets/Sensirion_Gas_Sensors_Datasheet_SGP30.pdf>`_).
    - Standardises the name used for logs and readings to "SGP30".

The installation and configuration is the same as the generic wrapper: ::

    # install the sensor library
    pip3 install adafruit-circuitpython-sgp30

    # create the sensor object
    import board
    i2c = board.I2C()
    sgp30 = adafruit_sgp30.Adafruit_SGP30(i2c)

    # prepare to poll / scrape
    SGP30Sensor(sgp30)

``SGP30Sensor`` polls at the default period of :mod:`GenericSensor <snsary.contrib.adafruit.generic>`. The `datasheet <https://www.sensirion.com/fileadmin/user_upload/customers/sensirion/Dokumente/9_Gas_Sensors/Datasheets/Sensirion_Gas_Sensors_Datasheet_SGP30.pdf>`_ says it should be 1s, but this doesn't work in practice, especially when the I2C bus is busy.

Humidity compensation
=====================

``SGP30Sensor`` is also an :mod:`Output <snsary.outputs.output>`. The TVOC and eCO2 readings from the SGP30 sensor need to be adjusted based on the absolute humidity of the surrounding air, which `can be calculated <https://www.sensirion.com/fileadmin/user_upload/customers/sensirion/Dokumente/9_Gas_Sensors/Datasheets/Sensirion_Gas_Sensors_Datasheet_SGP30.pdf>`_ from the "temperature" and "relative_humidity" in a batch of :mod:`Readings <snsary.models.reading>`: ::

    # SCD30 outputs the required readings
    GenericSensor(scd30).stream.tee(sgp30)

The required names - "tempeature" and "relative_humidity" - may not match some sensors. One way to work around this is to rename each :mod:`Reading <snsary.models.reading>` on the fly: ::

    OtherSensor.stream.filter_names('temp').rename(to='temperature').into(sgp30)
    OtherSensor.stream.filter_names('humid').rename(to='relative_humidity').into(sgp30)

Persistent IAQ baselines
========================

``SGP30Sensor`` emits ``baseline_TVOC`` and ``baseline_eCO2`` values, which are a moving average of the best environmental conditions the sensor has encountered. Higher values are better (from manual observation). `Adafruit recommend exposing the sensor to fresh air for at least 10 minutes <https://learn.adafruit.com/adafruit-sgp30-gas-tvoc-eco2-mox-sensor/circuitpython-wiring-test>`_ as part of configuring the IAQ baseline.

Every time the SGP30 sensor is sampled, the values of both baseline :mod:`Readings <snsary.models.reading>` are checked. Higher values are set as the new baseline for future readings; use ``persistent_baselines=False`` in the constructor to disable this. The baseline values are kept in persistent :mod:`storage <snsary.utils.storage>` so they survive restarts. See the :mod:`storage <snsary.utils.storage>` module for more details.

"""

from snsary.outputs import BatchOutput
from snsary.utils import storage

from .generic import GenericSensor


class SGP30Sensor(GenericSensor, BatchOutput):
    def __init__(self, device, persistent_baselines=True):
        BatchOutput.__init__(self)
        GenericSensor.__init__(self, device)

        if persistent_baselines:
            self.tracker = storage.MaxTracker(
                self.name,
                names=["baseline_TVOC", "baseline_eCO2"],
                on_change=self.baselines_changed,
            )
        else:
            self.tracker = storage.NullTracker()
            self.logger.debug("Persistent baselines disabled, ignoring.")

    @property
    def name(self):
        return "SGP30"

    def start(self):
        if self.tracker.values:
            self.baselines_changed(old={}, new=self.tracker.values)
        else:
            self.logger.debug("No baselines to restore, using defaults.")

        GenericSensor.start(self)

    def ready(self, elapsed_seconds, **kwargs):
        return elapsed_seconds > 15

    def sample(self, **kwargs):
        readings = GenericSensor.sample(self, **kwargs)
        self.tracker.update(readings)
        return readings

    def baselines_changed(self, old, new):
        self.logger.debug(f"Setting baselines: {new}")

        self.device.set_iaq_baseline(
            eCO2=new["baseline_eCO2"], TVOC=new["baseline_TVOC"]
        )

    def publish_batch(self, readings):
        temperatures = self.__filter(readings, "temperature")
        relative_humidities = self.__filter(readings, "relative_humidity")

        if not temperatures or not relative_humidities:
            self.logger.warning("Incomplete data for self-calibration.")
            return

        self.device.set_iaq_relative_humidity(
            celsius=temperatures[-1].value,
            relative_humidity=relative_humidities[-1].value,
        )

    def __filter(self, readings, name):
        return [reading for reading in readings if reading.name == name]



from datetime import datetime


class Reading:
    def __init__(self, *, sensor_name, name, timestamp, value):
        self.__sensor_name = sensor_name
        self.__name = name
        self.__value = value
        self.__timestamp = timestamp

    @property
    def sensor_name(self):
        return self.__sensor_name

    @property
    def name(self):
        return self.__name

    @property
    def value(self):
        return self.__value

    @property
    def timestamp(self):
        return self.__timestamp

    @property
    def datetime(self):
        return datetime.fromtimestamp(self.timestamp).astimezone()

    def __str__(self):
        return f"<{self.name} {self.timestamp} {self.value}>"

    def dup(self, **kwargs):
        source = dict(
            sensor_name=self.sensor_name,
            name=self.name,
            value=self.value,
            timestamp=self.timestamp,
        )

        return type(self)(**{**source, **kwargs})



from .mock_sensor import MockSensor
from .multi_source import MultiSource
from .polling_sensor import PollingSensor
from .sensor import Sensor
from .source import Source

__all__ = [
    "MockSensor",
    "PollingSensor",
    "Sensor",
    "Source",
    "MultiSource",
]



from snsary.utils import tracing

from .poller import Poller
from .sensor import Sensor


class PollingSensor(Sensor, Poller):
    start = Poller.start

    def __init__(self, *, period_seconds):
        Sensor.__init__(self)
        Poller.__init__(self, period_seconds=period_seconds)

    @tracing.capture_exceptions("snsary.sources.polling_sensor")
    def tick(self, **kwargs):
        readings = list(self.sample(**kwargs))
        self.logger.info(f"Collected {len(readings)} readings.")

        for reading in readings:
            self.stream.publish(reading)

    def sample(self, **kwargs):
        raise NotImplementedError()



"""
Outputs KWh consumption from a specified Smart Meter in `half hour intervals <https://developer.octopus.energy/docs/api/#consumption>`_. At the time of writing, data is only available for the previous day sometime on the following day (not sure exactly when). Create an instance with ``.from_env()``, which expects:

- OCTOPUS_MPAN
- OCTOPUS_SERIAL
- OCTOPUS_TOKEN
"""

import os
from datetime import timedelta

import pyrfc3339

from snsary.models import Reading
from snsary.sources import PollingSensor
from snsary.utils import request


class OctopusSensor(PollingSensor):
    CONSUMPTION_URL = "https://api.octopus.energy/v1/electricity-meter-points/{mpan}/meters/{serial_number}/consumption/?period_from={period_from}&order_by=period"

    @classmethod
    def from_env(cls):
        return cls(
            mpan=os.environ["OCTOPUS_MPAN"],
            serial_number=os.environ["OCTOPUS_SERIAL"],
            token=os.environ["OCTOPUS_TOKEN"],
        )

    def __init__(self, *, mpan, serial_number, token):
        PollingSensor.__init__(
            self,
            period_seconds=30 * 60,  # 30 mins
        )

        self.__token = token
        self.__mpan = mpan
        self.__serial_number = serial_number

    @property
    def name(self):
        return "octopus"

    def sample(self, now, **kwargs):
        start = now - timedelta(days=1)
        start = start.replace(hour=0, minute=0, second=0, microsecond=0)

        url = self.CONSUMPTION_URL.format(
            mpan=self.__mpan,
            serial_number=self.__serial_number,
            period_from=pyrfc3339.generate(start),
        )

        self.logger.debug(f"Request {url}")
        response = request.retrying_session().get(
            url,
            auth=(self.__token, ""),
        )
        response.raise_for_status()

        samples = response.json()["results"]
        self.logger.debug(f"Response {samples}")

        for sample in samples:
            yield self.__reading_from_sample(sample)

    def __reading_from_sample(self, sample):
        sample_datetime = pyrfc3339.parse(sample["interval_end"])
        sample_timestamp = int(sample_datetime.timestamp())

        return Reading(
            sensor_name=self.name,
            name="consumption",
            timestamp=sample_timestamp,
            value=sample["consumption"],
        )



import logging
from contextlib import contextmanager
from threading import Thread

import time_machine

from snsary import system
from snsary.outputs import MockOutput
from snsary.sources import MockSensor, MultiSource
from tests.conftest import retry


@contextmanager
def tmp_app(*, sensors=[], outputs=[]):
    try:
        MultiSource(*sensors).stream.into(*outputs)
        system.start()
        yield
        system.stop()
    finally:
        system.clear_services()


def test_system(caplog):
    caplog.set_level(logging.INFO)
    sensors = [MockSensor(period_seconds=1)]
    outputs = [MockOutput(), MockOutput(index=1)]

    def first_assertions():
        assert "INFO - [snsary] Started." in caplog.text
        assert "INFO - [snsary.mocksensor-0] Collected 1 readings." in caplog.text
        assert "INFO - [snsary.mockoutput-0] Reading: <abc 1650885071 0>" in caplog.text
        assert "INFO - [snsary.mockoutput-1] Reading: <abc 1650885071 0>" in caplog.text

    def second_assertions():
        assert "INFO - [snsary.mockoutput-0] Reading: <abc 1650885072 1>" in caplog.text
        assert "INFO - [snsary.mockoutput-1] Reading: <abc 1650885072 1>" in caplog.text

    def end_assertions():
        assert "INFO - [snsary] Stopping." in caplog.text
        assert "INFO - [snsary] Bye." in caplog.text
        assert "ERROR" not in caplog.text

    with time_machine.travel("2022-04-25T12:11:11+01:00", tick=False) as frozen_time:
        with tmp_app(sensors=sensors, outputs=outputs):
            retry(first_assertions)
            frozen_time.shift(1)
            retry(second_assertions)

    retry(end_assertions)


def test_failing_sensor(caplog):
    sensors = [MockSensor(fail=True, period_seconds=1)]

    def assertions():
        assert "ERROR - [snsary.mocksensor-0] problem-1" in caplog.text
        assert "ERROR - [snsary.mocksensor-0] problem-2" in caplog.text

    with tmp_app(sensors=sensors):
        retry(assertions)


def test_failing_output(caplog):
    caplog.set_level(logging.INFO)
    sensors = [MockSensor(period_seconds=1)]
    outputs = [MockOutput(fail=True)]

    def assertions():
        assert "INFO - [snsary.mocksensor-0] Collected 1 readings." in caplog.text
        assert "ERROR - [snsary.mockoutput-0] problem-1" in caplog.text
        assert "ERROR - [snsary.mockoutput-0] problem-2" in caplog.text

    with tmp_app(sensors=sensors, outputs=outputs):
        retry(assertions)


def test_stuck_sensor_service(caplog):
    caplog.set_level(logging.INFO)
    sensors = [MockSensor(hang=True), MockSensor(index=1)]

    def assertions():
        assert "INFO - [snsary.mocksensor-1] Collected 1 readings." in caplog.text
        assert "INFO - [snsary.mocksensor-0] Collected 1 readings." not in caplog.text

    with tmp_app(sensors=sensors):
        retry(assertions)

    def end_assertions():
        assert "ERROR - [snsary.mocksensor-0] Failed to stop." in caplog.text

    retry(end_assertions)


def test_stuck_output_async(caplog):
    caplog.set_level(logging.INFO)
    sensors = [MockSensor()]
    outputs = [MockOutput(hang=True), MockOutput(index=1)]

    def assertions():
        assert "INFO - [snsary.mockoutput-1] Reading" in caplog.text
        assert "INFO - [snsary.mockoutput-0] Reading" not in caplog.text

    with tmp_app(sensors=sensors, outputs=outputs):
        retry(assertions)

    def end_assertions():
        assert "Bye." in caplog.text
        assert "ERROR" not in caplog.text

    retry(end_assertions)


def test_wait_stop():
    thread = Thread(
        target=lambda: system.wait(handle_signals=False),
        daemon=True,
    )

    thread.start()
    thread.join(timeout=0.1)
    assert thread.is_alive()

    system.stop()
    thread.join(timeout=0.1)
    assert not thread.is_alive()



from .backend import start, start_and_wait, stop, wait
from .service import Service, clear_services

__all__ = [
    "start",
    "start_and_wait",
    "stop",
    "wait",
    "Service",
    "clear_services",
]



import functools

from ..logging import get_logger
from .config import Config
from .registry import Registry
from .sample import Sample

_config = Config()
_registry = Registry(_config)


def reset():
    _config.reset()
    _registry.reset()


def configure(settings):
    for path, value in settings.items():
        _config.set(path, value)


def capture_exceptions(trace_id):
    def _factory(fn):
        @functools.wraps(fn)
        def _wrapper(*args, **kwargs):
            enabled = _config.get(f"{trace_id}.enabled", default=False)
            monitors = _registry.get(trace_id)

            if not enabled:
                return fn(*args, **kwargs)

            try:
                result = fn(*args, **kwargs)

                for monitor in monitors:
                    monitor.analyse(Sample.SUCCESS)

                return result
            except Exception as e:
                for monitor in monitors:
                    monitor.analyse(Sample.FAILURE)

                get_logger().warning(e)

        return _wrapper

    return _factory



import atexit
import os
import pickle

from cachetools import TTLCache, cached

from ..logging import get_logger


class HasStore:
    @property
    def store(self):
        """
        Shortcut to calling ``get_storage()`` directly.
        """
        return get_storage()


@cached({})
def get_storage():
    """
    Returns a memoised dict-like store.
    """
    return _get_storage_backend()


def _get_storage_backend():
    path = os.environ.get("STORAGE_PATH", "")

    if not path:
        get_logger().debug("Using memory store.")
        return dict()

    get_logger().debug(f"Using store {path}.")
    ttl = os.environ.get("STORAGE_TTL", 86400)

    get_logger().debug(f"Store TTL set to {ttl}.")
    return _get_file_backend(path, int(ttl))


def _get_file_backend(path, ttl):
    try:
        cache = pickle.load(open(path, "rb"))
    except Exception:
        get_logger().debug(f"{path} not found.")
        get_logger().debug("Creating new store.")
        cache = TTLCache(float("inf"), ttl)

    def atexit_handler():
        get_logger().debug(f"Writing store {path}.")
        pickle.dump(cache, open(path, "wb"))

    atexit.register(atexit_handler)
    return cache



from threading import Event

from snsary.models import Reading

from .polling_sensor import PollingSensor


class MockSensor(PollingSensor):
    def __init__(
        self,
        *,
        fail=False,
        hang=False,
        period_seconds=5,
        index=0,
    ):
        PollingSensor.__init__(
            self,
            period_seconds=period_seconds,
        )

        self.__hang = hang
        self.__fail = fail
        self.__failures = 0
        self.__index = index

    @property
    def name(self):
        return f"mocksensor-{self.__index}"

    def sample(
        self,
        now,  # unused here (implicit test for kwarg)
        start_time,  # unused (implicitly kwarg check)
        timestamp,
        elapsed_seconds,
    ):
        if self.__fail:
            self.__failures += 1
            raise RuntimeError(f"problem-{self.__failures}")

        if self.__hang:
            Event().wait()

        return [
            Reading(
                sensor_name=self.name,
                name="abc",
                timestamp=timestamp,
                value=elapsed_seconds,
            )
        ]



"""
A :mod:`Stream <snsary.streams.stream>` is an :mod:`Output <snsary.outputs.output>` and a :mod:`Source <snsary.sources.source>` that provides some kind of additional functionality over connecting :mod:`Sources <snsary.sources.source>` to :mod:`Outputs <snsary.outputs.output>` directly. All :mod:`Sensors <snsary.sources.sensor>` expose a ``stream`` of their :mod:`Readings <snsary.models.reading>` e.g. ``MockSensor().stream`` returns an :mod:`AsyncStream <snsary.streams.async_stream>` to cope with flakey :mod:`Outputs <snsary.outputs.output>`.

Streams make it easy to subscribe multiple :mod:`Outputs <snsary.outputs.output>` with ``into``: ::

    # same as calling "subscribe" for each
    stream.into(MockOutput(), MockOutput())

Any :mod:`Stream <snsary.streams.stream>` can also be wrapped in a :mod:`Filter <snsary.functions.filter>` function e.g. ::

    # only output readings for sensor "foo"
    stream.apply(Filter.sensor_name('foo')).into(MockOutput())

    # only output readings called "a", etc.
    # convenience shortcut method for "apply"
    stream.filter_names('a', 'b', 'c').subscribe(MockOutput())

A stream can actually be used to ``apply`` any one-to-many :mod:`function <snsary.functions>` to the readings that pass through it. For example, to average the distinct readings received in a window: ::

    # outputs an average value every 3 seconds
    # for each distinct sensor / reading name;
    # convenience shortcut method for "apply"
    stream.average(seconds=3).into(MockOutput())
"""


from .async_stream import AsyncStream
from .simple_stream import SimpleStream
from .stream import Stream

__all__ = [
    "SimpleStream",
    "AsyncStream",
    "Stream",
]



"""
Represents a FIFO list of :mod:`Samples <snsary.utils.tracing.sample>`. :mod:`Samples <snsary.utils.tracing.sample>` are added to the list with ``add``. The oldest :mod:`Sample <snsary.utils.tracing.sample>` is ejected when the list reaches the specified ``max_length``.
"""


class History:
    def __init__(self, *, max_length):
        self.__max_length = max_length
        self.__backend = []

    @property
    def max_length(self):
        return self.__max_length

    def add(self, sample):
        self.__backend.append(sample)

        if len(self) > self.max_length:
            self.__backend.pop(0)

    def reset(self):
        self.__backend.clear()

    def __len__(self):
        return len(self.__backend)

    def __iter__(self):
        return iter(self.__backend)



from ..logging import HasLogger
from .backend import HasStore


class Tracker(HasLogger, HasStore):
    def __init__(self, id, **trackers):
        self.__trackers = trackers
        self.__id = id

    @property
    def values(self):
        return self.store.get(f"{self.__id}-tracked-values", {})

    def on_change(self, *, old, new):
        pass

    def update(self, readings):
        values = self.values
        new_values = {}

        for name, fn in self.__trackers.items():
            current_value = self.__filter_value(readings, name)

            if not current_value:
                self.logger.debug(f"Tracker missing value for {name}.")
                return

            if name not in values:
                self.logger.debug(f"Tracker filling value for {name}.")
                new_values[name] = current_value
            else:
                new_values[name] = fn(current_value, values[name])

        if new_values == values:
            self.logger.debug("Tracked values unchanged, continuing.")
            return

        self.logger.debug(f"Storing tracked values: {new_values}")
        self.store[f"{self.__id}-tracked-values"] = new_values
        self.on_change(old=values, new=new_values)

    def __filter_value(self, readings, name):
        for reading in readings:
            if reading.name == name:
                return reading.value



from .source import Source


class Sensor(Source):
    def __init__(self):
        from snsary.streams import AsyncStream

        self.__stream = AsyncStream()

    @property
    def stream(self):
        return self.__stream

    def subscribe(self, output):
        self.stream.subscribe(output)



"""
Sends batches of :mod:`Readings <snsary.models.reading>` as rows to Google BigQuery to be stored in a dataset / table called ``snsary.readings``, using `the Google BigQuery Storage API <https://github.com/googleapis/python-bigquery-storage>`_.

Create an instance with ``.from_env()``, which expects:

- GOOGLE_APPLICATION_CREDENTIALS - the path to `a JSON credentials file <https://cloud.google.com/iam/docs/creating-managing-service-account-keys#creating>`_
- GOOGLE_BIGQUERY_STREAM - of the form ``projects/<your_project_id>/datasets/snsary/tables/readings/streams/_default``

Setting up BigQuery
===================

You can use the BigQuery UI to do most of the setup.

1. Create a dataset called ``snsary``.

    - Do not enable table expiration (this is different to partition expiration).

2. Create a table called ``readings``.

    - Add columns ``timestamp``, ``host``, ``sensor``, ``metric`` and ``value``.
    - Use ``TIMESTAMP`` for ``timestamp``, ``FLOAT`` for ``value`` and otherwise ``STRING``.
    - Partition the table **by day** using values of the **timestamp** column.

3. Set up partition expiration e.g. ::

       ALTER TABLE snsary.readings
       SET OPTIONS (
        partition_expiration_days=425
       )

You will also need to create `a Google Cloud service account <https://cloud.google.com/iam/docs/service-accounts>`_ and corresponding API key. The service account should have the "BigQuery Data Editor" role or similar.

Querying the data
=================

Example query for data in the table: ::

    SELECT $__timeGroup(timestamp,$__interval), sensor, metric, avg(value)
    FROM `snsary.readings`
    where $__timeFilter(timestamp)
    group by $__timeGroup(timestamp,$__interval), sensor, metric
    order by 1 asc

Note that the ``$__`` functions are `defined by Grafana <https://grafana.com/grafana/plugins/grafana-bigquery-datasource/>`_. A service account reading the data will need to have "BigQuery Data Viewer" and "BigQuery Job User" roles.
"""
import os
import platform

import pytz
from google.api_core.retry import Retry
from google.cloud import bigquery_storage_v1
from google.cloud.bigquery_storage_v1 import types
from google.protobuf import descriptor_pb2

from snsary.outputs import BatchOutput

from . import reading_pb2


class BigQueryOutput(BatchOutput):
    RETRY_DEADLINE = 10

    def __init__(self, stream):
        BatchOutput.__init__(self)
        self.__stream = stream
        self.__client = bigquery_storage_v1.BigQueryWriteClient()

    @classmethod
    def from_env(cls):
        return cls(
            stream=os.environ["GOOGLE_BIGQUERY_STREAM"],
        )

    def publish_batch(self, readings):
        proto_readings = [self.__proto_row(reading) for reading in readings]
        self.__proto_send(proto_readings)

    def __proto_row(self, reading):
        """
        Copy a :mod:`Reading <snsary.models.reading>` into an intermediary protobuf model `generated from a schema <https://developers.google.com/protocol-buffers/docs/pythontutorial>`_.
        """
        return reading_pb2.Reading(
            # using UTC ensures test stability when run in different zones
            timestamp=reading.datetime.astimezone(pytz.utc).isoformat(),
            host=platform.node(),
            sensor=reading.sensor_name,
            metric=reading.name,
            value=reading.value,
        )

    def __proto_send(self, proto_readings):
        """
        Serialise and send a set of intermediary protobuf readings together with their schema, which is redundant after the first attempt but simpler to keep sending anyway. This method was written by following the `BigQuery Storage API reference <https://cloud.google.com/bigquery/docs/reference/storage/rpc/google.cloud.bigquery.storage.v1#google.cloud.bigquery.storage.v1.AppendRowsRequest>`_ and `an example implementation <https://github.com/googleapis/python-bigquery-storage/blob/main/samples/snippets/append_rows_pending.py>`_.
        """
        proto_descriptor = descriptor_pb2.DescriptorProto()
        reading_pb2.Reading.DESCRIPTOR.CopyToProto(proto_descriptor)

        proto_schema = types.ProtoSchema()
        proto_schema.proto_descriptor = proto_descriptor

        proto_rows = types.ProtoRows()

        for proto_reading in proto_readings:
            proto_rows.serialized_rows.append(
                proto_reading.SerializeToString(),
            )

        proto_data = types.AppendRowsRequest.ProtoData()
        proto_data.rows = proto_rows
        proto_data.writer_schema = proto_schema

        request = types.AppendRowsRequest()
        request.write_stream = self.__stream
        request.proto_rows = proto_data

        self.__client.append_rows(
            requests=iter([request]),
            retry=Retry(deadline=self.RETRY_DEADLINE),
        )



import logging
import sys
from threading import current_thread, main_thread

rootLogger = logging.getLogger("snsary")
loggers = dict()


class HasLogger:
    @property
    def name(self):
        return type(self).__name__

    @property
    def logger(self):
        return get_logger(self.name.lower())


def get_logger(name=None):
    thread_id = current_thread().ident

    if name:
        sublogger = logging.getLogger(f"snsary.{name}")
        loggers[thread_id] = sublogger
        return sublogger

    if main_thread().ident == thread_id:
        return rootLogger

    if thread_id in loggers:
        return loggers[thread_id]

    return logging.getLogger(f"snsary.anon-{thread_id}")


def configure_logging(level=logging.INFO):
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s - [%(name)s] %(message)s",
    )



"""
Tracing utility with a ``capture_exceptions`` decorator that instruments a given function to monitor success and failure (exceptions) over time; exceptions are captured and logged as warnings. Basic usage of the tracing utility looks like this (with global configuration): ::

    @tracing.capture_exceptions("some.unique.path")
    def noisy_error_function():
        ...

    tracing.configure({
        # tracing is disabled by default
        "enabled": True,

        # specify how to look for errors
        "monitors": [
            tracing.GapsAlert.factory(),
            tracing.LivenessAlert.factory(),
        ],

        # monitor each thread separately
        "thread_aware": True,
    })

With this config, the tracing system will create a new instance of each :mod:`Monitor <snsary.utils.tracing.monitors>` for each function being traced and each thread that calls the function. You can change this behaviour on a per-function basis. For example, customising ``some.unique.path`` looks like this: ::

    tracing.configure({
        "some.unique.path.monitors": [
            tracing.GapsMonitor.factory(
                max_gaps=5,
                history=SimpleHistory(max_length=10),
            )
        ]
    })


See the :mod:`Config <snsary.utils.tracing.config>` module for more details.
"""

from .config import Config
from .core import capture_exceptions, configure, reset
from .history import History
from .monitors import GapAlert, LivenessAlert, Monitor
from .registry import Registry
from .sample import Sample

__all__ = [
    "capture_exceptions",
    "configure",
    "GapAlert",
    "LivenessAlert",
    "Sample",
    "Monitor",
    "reset",
    "History",
    "Config",
    "Registry",
]



"""
A collection of tools to interact with a remote Grafana Cloud instance.

`GraphiteOutput` sends batches of :mod:`Readings <snsary.models.reading>` to Grafana Cloud, who provide `a custom ingest endpoint for Graphite metrics <https://grafana.com/docs/grafana-cloud/metrics-graphite/http-api/>`_. Metric names are of the form ``<prefix>.<sensor name>.<reading name>``. Using ``.from_env()`` to create an instance sets the ``<prefix>`` to the hostname of the machine. ``.from_env()`` expects ``GRAPHITE_URL`` to be set as an environment variable.
"""

import json
import os
import platform

from snsary.outputs import BatchOutput
from snsary.utils import request


class GraphiteOutput(BatchOutput):
    @classmethod
    def from_env(cls):
        return cls(
            url=os.environ["GRAPHITE_URL"],
            prefix=platform.node(),
        )

    def __init__(self, *, url, prefix):
        BatchOutput.__init__(self)
        self.__url = url
        self.__prefix = prefix

    def publish_batch(self, readings):
        data = self.__format(readings)
        self.logger.debug(f"Sending {data}")

        response = request.retrying_session().post(
            self.__url,
            data=data,
            headers={"Content-Type": "application/json"},
        )

        self.logger.debug(response.text)
        response.raise_for_status()

    def __format(self, readings):
        return json.dumps(
            [
                {
                    "name": f"{self.__prefix}.{reading.sensor_name}.{reading.name}",
                    "value": reading.value,
                    "time": reading.timestamp,
                    "interval": 1,
                }
                for reading in readings
            ]
        )



"""
Base class for functions that aggregate :mod:`Readings <snsary.models.reading>` over fixed, consecutive periods, specified as the keyword arguments for a ``timedelta`` (seconds, minutes, etc.).

A subclass should define an ``aggregate`` method that is called each time the age of the window for a sensor / reading name pair reaches the specified period.

The windows are consecutive, not moving: after a window has been aggregated a new window is started using the :mod:`Reading <snsary.models.reading>` that triggered the previous window to close.

Windows also act as a :mod:`Service <snsary.system.service>` in order to persist and restore :mod:`Readings <snsary.models.reading>` when the program they are used in starts and stops. See the :mod:`storage <snsary.utils.storage>` module for more details.
"""


from datetime import timedelta

from snsary import system

from .function import Function


class Window(Function, system.Service):
    def __init__(self, **kwargs):
        system.Service.__init__(self)
        self.__period = timedelta(**kwargs).total_seconds()
        self.__windows = dict()

    def stop(self):
        for key, window in self.__windows.items():
            self.logger.debug(f"Storing window for {key}.")
            self.store[key] = window

    def __call__(self, reading):
        key = self.key(reading)

        if key not in self.__windows and key in self.store:
            self.logger.debug(f"Restoring window for {key}.")
            self.__windows[key] = self.store[key]

        if key not in self.__windows:
            self.logger.debug(f"Starting window for {key}.")
            self.__windows[key] = [reading]
            return []

        readings = self.__windows[key]
        start = readings[0].timestamp
        age = reading.timestamp - start

        if age >= self.__period:
            self.logger.debug(f"Closing window for {key}.")
            self.__windows[key] = [reading]
            return self.aggregate(readings)

        readings.append(reading)
        return []

    def aggregate(self, readings):
        raise NotImplementedError()

    def key(self, reading):
        return f"window-{int(self.__period)}-{reading.sensor_name}-{reading.name}"



"""
Base class for anything that needs to perform an action during the ``start`` or ``stop`` phases of a program. This could be something like spawning a thread, or storing data to disk. Each instance of a service is recorded in a global registry so it can be managed by the :mod:`system <snsary.system>` module.
"""

from snsary.utils import logging, storage

_instances = []


class Service(logging.HasLogger, storage.HasStore):
    def __init__(self):
        global _instances
        _instances += [self]

    def start(self):
        """
        Called synchronously by the :mod:`system <snsary.system>` module as part of program startup.
        """
        pass

    def stop(self):
        """
        Called asynchronously by the :mod:`system <snsary.system>` module when a program is stopping. The execution will eventually time out to ensure the overall program stops in good time.
        """
        pass


def clear_services():
    """
    Clear the global list of service instances. For testing use only.
    """
    global _instances
    _instances = []


def get_services():
    """
    Return a list of all service instances.
    """
    return _instances



from types import MemberDescriptorType as member_descriptor

from .logging import get_logger


def extract_from(value, *, prefix):
    if isinstance(value, dict):
        for key, item in value.items():
            yield from extract_from(item, prefix=f"{prefix}-{key}")

    elif hasattr(value, "_asdict"):
        yield from extract_from(value._asdict(), prefix=prefix)

    elif isinstance(value, (tuple, list)):
        for index, item in enumerate(value):
            yield from extract_from(item, prefix=f"{prefix}-{index}")

    elif isinstance(value, int):
        yield (prefix, int(value))

    elif isinstance(value, float):
        yield (prefix, value)

    else:
        get_logger().debug(f"Ignoring {value} as not a number.")


class for_class:
    def __init__(self, klass):
        self.__props = [
            name
            for name, value in vars(klass).items()
            if isinstance(value, (property, member_descriptor))
            and not name.startswith("_")
        ]

    def __call__(self, instance):
        for prop in self.__props:
            value = getattr(instance, prop)
            yield from extract_from(value, prefix=prop)



from threading import Event

from .output import Output


class MockOutput(Output):
    def __init__(
        self,
        *,
        fail=False,
        hang=False,
        index=0,
    ):
        self.__fail = fail
        self.__hang = hang
        self.__failures = 0
        self.__index = index

    @property
    def name(self):
        return f"mockoutput-{self.__index}"

    def publish(self, reading):
        if self.__hang:
            Event().wait()

        if self.__fail:
            self.__failures += 1
            raise RuntimeError(f"problem-{self.__failures}")

        self.logger.info(f"Reading: {reading}")



import setuptools

extras_require = {
    "adafruit": [],
    "awair": [
        "requests>=2.28.2",
        "pyrfc3339>=1.1",
    ],
    "datastax": [
        "gql[requests]>=3.4.0",
    ],
    "google": [
        "google-cloud-bigquery-storage>=2.18.1",
    ],
    "grafana": [
        "requests>=2.28.2",
    ],
    "influxdb": [
        "influxdb-client>=1.36.0",
    ],
    "octopus": [
        "requests>=2.28.2",
        "pyrfc3339>=1.1",
    ],
    "pimoroni": [],
    "psutil": [
        "psutil>=5.9.4",
    ],
    "pypms": [
        "pypms>=0.7.1",
    ],
}

extras_require["all"] = [
    package for group in extras_require.values() for package in group
]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="snsary",
    version="0.0.1",
    author="Ben Thorner",
    author_email="benthorner@users.noreply.github.com",
    description="A framework for sensor metrics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/benthorner/snsary",
    project_urls={
        "Bug Tracker": "https://github.com/benthorner/snsary/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.9",
    install_requires=[
        "cachetools>=5.3.0",
        "wrapt>=1.14.1",
    ],
    extras_require=extras_require,
)



from snsary.functions import Filter, Rename, WindowAverage, WindowSummary
from snsary.outputs import Output
from snsary.sources import Source


class Stream(Source, Output):
    def into(self, *outputs):
        for output in outputs:
            self.subscribe(output)

    def tee(self, output):
        """
        Subscribes the specified :mod:`Output <snsary.outputs.output>` to the stream using ``into``, but unlike ``into`` also returns the stream for further use, similar to the Unix ``tee`` command.
        """
        self.into(output)
        return self

    def apply(self, function):
        from .func_stream import FuncStream

        return FuncStream(self, function)

    def filter_names(self, *names):
        return self.apply(Filter.names(*names))

    def average(self, **kwargs):
        """
        Returns a new stream that applies a :mod:`WindowAverage <snsary.functions.window_average>` to all :mod:`Readings <snsary.models.reading>` over a period, specified as the keyword arguments for a ``timedelta``.
        """
        return self.apply(WindowAverage(**kwargs))

    def summarize(self, **kwargs):
        """
        Returns a new stream that applies a :mod:`WindowSummary <snsary.functions.window_summary>` to all :mod:`Readings <snsary.models.reading>` over a period, specified as the keyword arguments for a ``timedelta``.
        """
        return self.apply(WindowSummary(**kwargs))

    def rename(self, **kwargs):
        """
        Returns a new stream that applies a :mod:`Rename <snsary.functions.rename>` to all :mod:`Readings <snsary.models.reading>`. For example, ``rename(append="foo")`` will append "foo" to all reading names.
        """
        return self.apply(Rename(**kwargs))



from .function import Function


class Filter(Function):
    def __init__(self, allow_fn=lambda _: True):
        self.__allow_fn = allow_fn

    def __call__(self, reading):
        return [reading] if self.__allow_fn(reading) else []

    @property
    def invert(self):
        return Filter(lambda reading: not self(reading))

    @classmethod
    def sensor_name(cls, name):
        return cls(lambda reading: reading.sensor_name == name)

    @classmethod
    def names(cls, *names):
        return cls(lambda reading: reading.name in names)



import psutil

from snsary.models import Reading
from snsary.sources import PollingSensor
from snsary.utils import scraper


class PSUtilSensor(PollingSensor):
    FUNCTIONS = {
        "cpu_times": {},
        "cpu_percent": {"interval": 1},
        "cpu_count": {},
        "cpu_stats": {},
        "getloadavg": {},
        "virtual_memory": {},
        "swap_memory": {},
        "disk_usage": {"path": "/"},
        "disk_io_counter": {},
        "net_io_counters": {},
        "sensors_temperatures": {},
        "sensors_fans": {},
        "sensors_battery": {},
    }

    def __init__(self, functions=FUNCTIONS):
        self.__functions = functions
        PollingSensor.__init__(self, period_seconds=10)

    @property
    def name(self):
        return "psutil"

    def sample(self, timestamp, **kwargs):
        for fname in self.__functions:
            yield from self.__readings_from_function(
                timestamp,
                fname,
            )

    def __readings_from_function(self, timestamp, fname):
        for name, value in self.__sample_fn(fname):
            yield Reading(
                sensor_name=self.name,
                name=name,
                value=value,
                timestamp=timestamp,
            )

    def __sample_fn(self, fname):
        if not hasattr(psutil, fname):
            self.logger.debug(f"Skipping {fname} as not available.")
            return []

        kwargs = self.__functions[fname]
        value = getattr(psutil, fname)(**kwargs)

        self.logger.debug(f"Scraping {fname} => {value}")
        return scraper.extract_from(value, prefix=fname)



"""
Outputs raw (unaveraged) data from all sensors of a specified device every 5 minutes. The long period is necessary due to the severe rate limiting Awair have on `their API <https://docs.developer.getawair.com/#local-api>`_.

Calling ``.discover()`` will auto-discover and return a list of AwairSensors associated with an account, which is more convenient than having to manually find and specify the details of each device. You can also call ``.discover_from_env``, which expects:

- AWAIR_TOKEN
"""

import os
from datetime import timedelta

import pyrfc3339

from snsary.models import Reading
from snsary.sources import PollingSensor
from snsary.utils import logging, request


class AwairSensor(PollingSensor):
    DEVICES_URL = "https://developer-apis.awair.is/v1/users/self/devices"
    DATA_URL = "https://developer-apis.awair.is/v1/users/self/devices/{deviceType}/{deviceId}/air-data/raw?from={from}&desc=False"
    PERIOD = 5 * 60  # 5 mins

    @classmethod
    def discover_from_env(cls):
        return cls.discover(
            token=os.environ["AWAIR_TOKEN"],
        )

    @classmethod
    def discover(cls, *, token):
        logger = logging.get_logger()
        logger.debug(f"Request {cls.DEVICES_URL}")
        response = request.retrying_session().get(
            cls.DEVICES_URL,
            headers={"Authorization": f"Bearer {token}"},
        )
        response.raise_for_status()
        devices = response.json()["devices"]

        logger.info(f"Discovered {len(devices)} Awair devices.")
        logger.debug(f"Discovered {devices}")

        return [
            AwairSensor(
                device_type=device["deviceType"],
                device_id=device["deviceId"],
                token=token,
            )
            for device in devices
        ]

    def __init__(self, *, device_type, device_id, token):
        PollingSensor.__init__(self, period_seconds=self.PERIOD)
        self.__token = token
        self.__device_type = device_type
        self.__device_id = device_id

    @property
    def name(self):
        return f"{self.__device_type}-{self.__device_id}"

    def sample(self, now, **kwargs):
        # subtract double period in case of delayed readings
        sample_start = now - timedelta(seconds=self.PERIOD * 2)

        url = self.DATA_URL.format(
            **{
                "deviceType": self.__device_type,
                "deviceId": self.__device_id,
                "from": pyrfc3339.generate(sample_start),
            }
        )

        self.logger.debug(f"Request {url}")
        response = request.retrying_session().get(
            url,
            headers={"Authorization": f"Bearer {self.__token}"},
        )
        response.raise_for_status()
        samples = response.json()["data"]
        self.logger.debug("Response {samples}")

        for sample in samples:
            yield from self.__readings_from_sample(sample)

    def __readings_from_sample(self, sample):
        sample_datetime = pyrfc3339.parse(sample["timestamp"])
        sample_timestamp = int(sample_datetime.timestamp())

        for sensor in sample["sensors"]:
            yield (
                Reading(
                    sensor_name=self.name,
                    name=sensor["comp"],
                    timestamp=sample_timestamp,
                    value=sensor["value"],
                )
            )
